import copy
import numpy as np


class PSO:
    def __init__(self, problem, swarm_size, dim, max_iter, minimize=True):
        '''
        Initialize the PSO algorithm.

        :param problem: The problem to be optimized.
        :param swarm_size: The size of the swarm.
        :param dim: The dimension of the problem.
        :param max_iter: The maximum number of iterations.
        :param minimize: Whether to minimize or maximize the problem.
        '''
        self.problem = problem
        self.swarm_size = swarm_size
        self.dim = dim
        self.max_iter = max_iter
        # TODO: deal with None max_iter
        # TODO: different possible initializations
        self.swarm = np.random.uniform(self.problem.bounds[:, 0],
                                       self.problem.bounds[:, 1],
                                       (self.swarm_size, self.dim))
        # TODO: different velocity initialization
        self.velocity = np.zeros((self.swarm_size, self.dim))
        self.minimize = minimize
        if not self.minimize:
            self.problem = lambda x: -self.problem(x)
        self.pbest = copy.deepcopy(self.swarm)
        self.pbest_fitness = self.problem(self.pbest)

    def update(self):
        '''
        Update the swarm for one iteration.
        '''
        pass

    def optimize(self):
        '''
        Optimize the problem.
        '''
        pass



import numpy as np


class BenchFunc:
    def __init__(self):
        self.bounds = None
        self.optimum_value = None
        self.name = None
        self.dim = None

    def eval(self, x):
        pass

    def __call__(self, x):
        return self.eval(x)

    def get_bounds(self):
        return self.bounds

    def get_optimum_value(self):
        return self.optimum_value

    def get_name(self):
        return self.name

    def get_dim(self):
        return self.dim


class Beale(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-4.5, 4.5], [-4.5, 4.5]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Beale"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return (1.5 - x[0] + x[0] * x[1]) ** 2 + (
                2.25 - x[0] + x[0] * x[1] ** 2) ** 2 + (
                       2.625 - x[0] + x[0] * x[1] ** 3) ** 2


class GoldsteinPrice(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-2, 2], [-2, 2]])
        self.optimum_value = 3.0
        self.dim = 2
        self.name = "Goldstein-Price"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return (1 + (x[0] + x[1] + 1) ** 2 * (
                19 - 14 * x[0] + 3 * x[0] ** 2 - 14 * x[1] + 6 * x[0] * x[
                    1] + 3 * x[1] ** 2)) * (30 + (2 * x[0] - 3 * x[1]) ** 2 * (
                        18 - 32 * x[0] + 12 * x[0] ** 2 + 48 * x[1] - 36 * x[0]
                        * x[1] + 27 * x[1] ** 2))


class Booth(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-10, 10], [-10, 10]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Booth"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return (x[0] + 2 * x[1] - 7) ** 2 + (2 * x[0] + x[1] - 5) ** 2


class BukinN6(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-15, -5], [-3, 3]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Bukin N.6"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return 100 * np.sqrt(np.abs(x[1] - 0.01 * x[0] ** 2)) + 0.01 * np.abs(
            x[0] + 10)


class Matyas(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-10, 10], [-10, 10]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Matyas"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return 0.26 * (x[0] ** 2 + x[1] ** 2) - 0.48 * x[0] * x[1]


class LevyN13(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-10, 10], [-10, 10]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Levi N.13"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return np.sin(3 * np.pi * x[0]) ** 2 + (x[0] - 1) ** 2 * (
                1 + np.sin(3 * np.pi * x[1]) ** 2) + (x[1] - 1) ** 2 * (
                       1 + np.sin(2 * np.pi * x[1]) ** 2)


class Himmelblau(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-5, 5], [-5, 5]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Himmelblau"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return (x[0] ** 2 + x[1] - 11) ** 2 + (x[0] + x[1] ** 2 - 7) ** 2


class ThreeHumpCamel(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-5, 5], [-5, 5]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Three-Hump Camel"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return 2 * x[0] ** 2 - 1.05 * x[0] ** 4 + x[0] ** 6 \
            / 6 + x[0] * x[1] + x[1] ** 2


class Easom(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-100, 100], [-100, 100]])
        self.optimum_value = -1.0
        self.dim = 2
        self.name = "Easom"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return -np.cos(x[0]) * np.cos(x[1]) * np.exp(
            -(x[0] - np.pi) ** 2 - (x[1] - np.pi) ** 2)


class CrossInTray(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-10, 10], [-10, 10]])
        self.optimum_value = -2.06261
        self.dim = 2
        self.name = "Cross-in-Tray"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return -0.0001 * (np.abs(np.sin(x[0]) * np.sin(x[1]) * np.exp(
            np.abs(100 - np.sqrt(x[0] ** 2 + x[1] ** 2) / np.pi))) + 1) ** 0.1


class Eggholder(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-512, 512], [-512, 512]])
        self.optimum_value = -959.6407
        self.dim = 2
        self.name = "Eggholder"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return -(x[1] + 47) * np.sin(np.sqrt(np.abs(x[1] + x[0] / 2 + 47))) \
               - x[0] * np.sin(np.sqrt(np.abs(x[0] - (x[1] + 47))))


class HolderTable(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-10, 10], [-10, 10]])
        self.optimum_value = -19.2085
        self.dim = 2
        self.name = "Holder Table"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return -np.abs(np.sin(x[0]) * np.cos(x[1]) * np.exp(
            np.abs(1 - np.sqrt(x[0] ** 2 + x[1] ** 2) / np.pi)))


class McCormick(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-1.5, 4], [-3, 4]])
        self.optimum_value = -1.9133
        self.dim = 2
        self.name = "McCormick"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return np.sin(x[0] + x[1]) + (x[0] - x[1]) ** 2 - 1.5 * x[0] + 2.5 * x[
            1] + 1


class SchafferN2(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-100, 100], [-100, 100]])
        self.optimum_value = 0.0
        self.dim = 2
        self.name = "Schaffer N.2"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return 0.5 + (np.sin(x[0] ** 2 - x[1] ** 2) ** 2 - 0.5) / (
                1 + 0.001 * (x[0] ** 2 + x[1] ** 2)) ** 2


class SchafferN4(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-100, 100], [-100, 100]])
        self.optimum_value = 0.292579
        self.dim = 2
        self.name = "Schaffer N.4"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return 0.5 + (np.cos(
            np.sin(np.abs(x[0] ** 2 - x[1] ** 2))) ** 2 - 0.5) / (
                       1 + 0.001 * (x[0] ** 2 + x[1] ** 2)) ** 2


class DropWave(BenchFunc):
    def __init__(self):
        super().__init__()
        self.bounds = np.array([[-5.12, 5.12], [-5.12, 5.12]])
        self.optimum_value = -1.0
        self.dim = 2
        self.name = "Drop-Wave"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return -(1 + np.cos(12 * np.sqrt(x[0] ** 2 + x[1] ** 2))) / \
            (0.5 * (x[0] ** 2 + x[1] ** 2) + 2)


# TODO: Fix to work with broadcasting
# class Shubert(BenchFunc):
#     def __init__(self):
#         super().__init__()
#         self.bounds = np.array([[-10, 10], [-10, 10]])
#         self.optimum_value = -186.7309
#         self.dim = 2
#         self.name = "Shubert"
#         self.five_range = np.arange(1, 6)
#
#     def eval(self, x):
#         if not isinstance(x, np.ndarray):
#             raise TypeError("Input (x) must be a numpy array.")
#         if x.shape[-1] != self.dim:
#             raise ValueError("Input (x) must
#             have shape ({},).".format(self.dim))
#         x = np.transpose(x)
#         return np.sum(self.five_range * np.cos((self.five_range + 1) * x[0] +
#         self.five_range)) * np.sum(self.five_range *
#         np.cos((self.five_range + 1) * x[1] + self.five_range))


class VaryingDimBenchFunc(BenchFunc):
    def __init__(self, dim):
        super().__init__()
        if not isinstance(dim, int):
            raise TypeError("Dimension (dim) must be an integer.")
        if dim <= 1:
            raise ValueError("Dimension (dim) must be greater than 1.")
        self.dim = dim


class Rastrigin(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-5.12, 5.12], A=10):
        super().__init__(dim)
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.A = A
        self.name = "Rastrigin"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return np.sum((x ** 2) - (self.A * np.cos(2 * np.pi * x)), axis=0) + (
                self.A * self.dim)


class Ackley(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-32.768, 32.768], a=20, b=0.2,
                 c=2 * np.pi):
        super().__init__(dim)
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.a = a
        self.b = b
        self.c = c
        self.name = "Ackley"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return -self.a * np.exp(
            -self.b * np.sqrt(np.sum(x ** 2, axis=0) / self.dim)) - np.exp(
            (1 / self.dim) * np.sum(np.cos(self.c * x),
                                    axis=0)) + self.a + np.exp(1)


class Sphere(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-5.12, 5.12]):
        super().__init__(dim)
        if not isinstance(bounds, list) or len(bounds) != 2:
            raise TypeError("Bounds (bounds) must be a list of length 2.")
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.name = "Sphere"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return np.sum(x ** 2, axis=0)


class Rosenbrock(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-10, 10]):
        super().__init__(dim)
        if not isinstance(bounds, list) or len(bounds) != 2:
            raise TypeError("Bounds (bounds) must be a list of length 2.")
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.name = "Rosenbrock"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return np.sum(100 * (x[1:] - x[:-1] ** 2) ** 2 + (1 - x[:-1]) ** 2,
                      axis=0)


class StyblinskiTang(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-5, 5]):
        super().__init__(dim)
        if not isinstance(bounds, list) or len(bounds) != 2:
            raise TypeError("Bounds (bounds) must be a list of length 2.")
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = -39.16616570377142 * dim
        self.name = "StyblinskiTang"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return np.sum(x ** 4 - 16 * x ** 2 + 5 * x, axis=0) / 2


class Griewank(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-600, 600]):
        super().__init__(dim)
        if not isinstance(bounds, list) or len(bounds) != 2:
            raise TypeError("Bounds (bounds) must be a list of length 2.")
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.name = "Griewank"
        self.range_sqrt = np.sqrt(np.arange(1, self.dim + 1))

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return 1 + np.sum(x ** 2, axis=0) / 4000 - np.prod(
            np.cos(x / self.range_sqrt), axis=0)


class Levy(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-10, 10]):
        super().__init__(dim)
        if not isinstance(bounds, list) or len(bounds) != 2:
            raise TypeError("Bounds (bounds) must be a list of length 2.")
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.name = "Levy"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        w = 1 + (x - 1) / 4
        return np.sin(np.pi * w[0]) ** 2 + np.sum(
            (w[:-1] - 1) ** 2 * (1 + 10 * np.sin(np.pi * w[:-1] + 1) ** 2),
            axis=0) + (w[-1] - 1) ** 2 * (1 + np.sin(2 * np.pi * w[-1]) ** 2)


# TODO: Fix Schwefel
# class Schwefel(VaryingDimBenchFunc):
#     def __init__(self, dim, bounds=[-500, 500]):
#         super().__init__(dim)
#         if not isinstance(bounds, list) or len(bounds) != 2:
#             raise TypeError("Bounds (bounds) must be a list of length 2.")
#         self.bounds = np.array([bounds] * dim)
#         self.optimum_value = 0.0
#         self.name = "Schwefel"
#
#     def eval(self, x):
#         if not isinstance(x, np.ndarray):
#             raise TypeError("Input (x) must be a numpy array.")
#         if x.shape[-1] != self.dim:
#             raise ValueError("Input (x) must
#             have shape ({},).".format(self.dim))
#         x = np.transpose(x)
#         return 418.9829 * self.dim - np.sum(x * np.sin(np.sqrt(np.abs(x))),
#         axis=0)


# TODO: Fix Trid
# class Trid(VaryingDimBenchFunc):
#     def __init__(self, dim):
#         super().__init__(dim)
#         self.bounds = np.array([[-(dim ** 2), dim ** 2]] * dim)
#         self.optimum_value = -dim * (dim + 4) * (dim - 1) / 6
#         self.name = "Trid"
#
#     def eval(self, x):
#         if not isinstance(x, np.ndarray):
#             raise TypeError("Input (x) must be a numpy array.")
#         if x.shape[-1] != self.dim:
#             raise ValueError("Input (x) must have shape
#             ({},).".format(self.dim))
#         x = np.transpose(x)
#         return np.sum((x - 1) ** 2, axis=0) - np.sum(x[:-1] * x[1:], axis=0)


class DixonPrice(VaryingDimBenchFunc):
    def __init__(self, dim, bounds=[-10, 10]):
        super().__init__(dim)
        self.bounds = np.array([bounds] * dim)
        self.optimum_value = 0.0
        self.name = "DixonPrice"

    def eval(self, x):
        if not isinstance(x, np.ndarray):
            raise TypeError("Input (x) must be a numpy array.")
        if x.shape[-1] != self.dim:
            raise ValueError(
                "Input (x) must have shape ({},).".format(self.dim))
        x = np.transpose(x)
        return (x[0] - 1) ** 2 + np.sum(2 * (2 * np.arange(1, self.dim) + 1)
                                        * (2 * x[1:] ** 2 - x[:-1]) ** 2,
                                        axis=0)



from metaoptim.pso.pso import PSO
import copy
import numpy as np
import numba
from tqdm import tqdm
import metaoptim.config as config

numba.config.DISABLE_JIT = config._disable_jit
numba.config.NUMBA_NUM_THREADS = config._numba_num_threads


@numba.jit(nopython=config._numba_nopython, cache=config._numba_cache,
           parallel=config._numba_parallel, nogil=config._numba_nogil)
def _update_velocity_helper(w, c1, c2, swarm, velocity, pbest, gbest):
    """
    Helper function for updating the velocity.

    :param w: The inertia weight.
    :param c1: The cognitive parameter.
    :param c2: The social parameter.
    :param swarm: The swarm.
    :param velocity: The velocity.
    :param pbest: The personal best.
    :param gbest: The global best.
    :return: The updated velocity.
    """
    r1 = np.random.uniform(0, 1, swarm.shape)
    r2 = np.random.uniform(0, 1, swarm.shape)
    return w * velocity + c1 * r1 * (pbest - swarm) + c2 * r2 * (gbest - swarm)


@numba.jit(nopython=config._numba_nopython, cache=config._numba_cache,
           parallel=config._numba_parallel, nogil=config._numba_nogil)
def _update_swarm_helper(swarm, velocity):
    '''
    Helper function for updating the swarm.

    :param swarm: The swarm.
    :param velocity: The velocity.
    :return: The updated swarm.
    '''
    return swarm + velocity


@numba.jit(nopython=config._numba_nopython, cache=config._numba_cache,
           parallel=config._numba_parallel, nogil=config._numba_nogil)
def _update_bests_helper(fitness, swarm, pbest, pbest_fitness, gbest,
                         gbest_fitness, minimize):
    """
    Helper function for updating the personal and global bests.

    :param fitness: The fitness of the swarm.
    :param swarm: The swarm.
    :param pbest: The personal best.
    :param pbest_fitness: The fitness of the personal best.
    :param gbest: The global best.
    :param gbest_fitness: The fitness of the global best.
    :param minimize: Whether to minimize or maximize the problem.
    :return: The updated personal best, the updated personal best fitness, the
    updated global best, and the updated global best fitness.
    """
    if minimize:
        pbest_mask = fitness < pbest_fitness
    else:
        pbest_mask = fitness > pbest_fitness
    pbest[pbest_mask] = swarm[pbest_mask]
    pbest_fitness[pbest_mask] = fitness[pbest_mask]
    if minimize:
        best_particle = np.argmin(pbest_fitness)
    else:
        best_particle = np.argmax(pbest_fitness)
    if minimize:
        if pbest_fitness[best_particle] < gbest_fitness:
            gbest = np.copy(pbest[best_particle])
            gbest_fitness = pbest_fitness[best_particle]
    else:
        if pbest_fitness[best_particle] > gbest_fitness:
            gbest = np.copy(pbest[best_particle])
            gbest_fitness = pbest_fitness[best_particle]
    return pbest, pbest_fitness, gbest, gbest_fitness


class GBestPSO(PSO):
    def __init__(self, problem, swarm_size, dim, max_iter, minimize=True,
                 w=0.7298, c1=1.49618, c2=1.49618):
        """
        A class for the global best particle swarm optimization algorithm. This
        particle swarm optimization algorithm uses an inertia weight, cognitive
        parameter, and social parameter.

        :param problem: The problem to be optimized.
        :param swarm_size: The size of the swarm.
        :param dim: The dimension of the swarm.
        :param max_iter: The maximum number of iterations.
        :param minimize: Whether to minimize the problem.
        :param w: The inertia weight.
        :param c1: The cognitive parameter.
        :param c2: The social parameter.
        """
        super().__init__(problem, swarm_size, dim, max_iter, minimize)
        self.gbest = copy.deepcopy(self.swarm[np.argmin(self.pbest_fitness)])
        self.gbest_fitness = self.problem(self.gbest)
        self.w = w
        self.c1 = c1
        self.c2 = c2
        # TODO: velocity clamping

    def update(self):
        """
        Update the swarm for one iteration.
        """
        self._update_bests()
        self._update_velocity()
        self._update_swarm()

    def optimize(self):
        """
        Optimize the problem.
        """
        for _ in tqdm(range(self.max_iter)):
            self.update()
        return self.gbest, self.gbest_fitness

    def _update_bests(self):
        """
        Update the best positions and fitnesses.
        """
        fitness = self.problem(self.swarm)
        self.pbest, self.pbest_fitness, self.gbest, self.gbest_fitness = \
            _update_bests_helper(fitness, self.swarm, self.pbest,
                                 self.pbest_fitness, self.gbest,
                                 self.gbest_fitness, self.minimize)

    def _update_velocity(self):
        """
        Update the velocity of the swarm.
        """
        self.velocity = _update_velocity_helper(self.w, self.c1, self.c2,
                                                self.swarm, self.velocity,
                                                self.pbest, self.gbest)

    def _update_swarm(self):
        '''
        Update the swarm.
        '''
        self.swarm = _update_swarm_helper(self.swarm, self.velocity)



import inspect
import sys

from metaoptim.bench_func import Beale, GoldsteinPrice, Booth, BukinN6, \
    Matyas, LevyN13, Himmelblau, ThreeHumpCamel, Easom, CrossInTray, \
    Eggholder, McCormick, SchafferN2, SchafferN4, DropWave, Rastrigin, \
    Rosenbrock, Ackley, Sphere, StyblinskiTang, Griewank, Levy, DixonPrice
import unittest
import numpy as np
from numpy.testing import assert_array_almost_equal


class TestBeale(unittest.TestCase):
    def test(self):
        beale = Beale()
        x = np.array([3, 0.5])
        self.assertAlmostEqual(beale(x), 0.0)

    def test_multiple(self):
        beale = Beale()
        x = np.array([[3, 0.5], [3, 0.5]])
        assert_array_almost_equal(beale(x), np.array([0.0, 0.0]))


class TestGoldsteinPrice(unittest.TestCase):
    def test(self):
        goldstein_price = GoldsteinPrice()
        x = np.array([0, -1])
        self.assertAlmostEqual(goldstein_price(x), 3.0)

    def test_multiple(self):
        goldstein_price = GoldsteinPrice()
        x = np.array([[0, -1], [0, -1]])
        assert_array_almost_equal(goldstein_price(x), np.array([3.0, 3.0]))


class TestBooth(unittest.TestCase):
    def test(self):
        booth = Booth()
        x = np.array([1, 3])
        self.assertAlmostEqual(booth(x), 0.0)

    def test_multiple(self):
        booth = Booth()
        x = np.array([[1, 3], [1, 3]])
        assert_array_almost_equal(booth(x), np.array([0.0, 0.0]))


class TestBukinN6(unittest.TestCase):
    def test(self):
        bukin_n6 = BukinN6()
        x = np.array([-10, 1])
        self.assertAlmostEqual(bukin_n6(x), 0.0)

    def test_multiple(self):
        bukin_n6 = BukinN6()
        x = np.array([[-10, 1], [-10, 1]])
        assert_array_almost_equal(bukin_n6(x), np.array([0.0, 0.0]))


class TestMatyas(unittest.TestCase):
    def test(self):
        matyas = Matyas()
        x = np.array([0, 0])
        self.assertAlmostEqual(matyas(x), 0.0)

    def test_multiple(self):
        matyas = Matyas()
        x = np.array([[0, 0], [0, 0]])
        assert_array_almost_equal(matyas(x), np.array([0.0, 0.0]))


class TestLevyN13(unittest.TestCase):
    def test(self):
        levy_n13 = LevyN13()
        x = np.array([1, 1])
        self.assertAlmostEqual(levy_n13(x), 0.0)

    def test_multiple(self):
        levy_n13 = LevyN13()
        x = np.array([[1, 1], [1, 1]])
        assert_array_almost_equal(levy_n13(x), np.array([0.0, 0.0]))


class TestHimmelblau(unittest.TestCase):
    def test(self):
        himmelblau = Himmelblau()
        x = np.array([3, 2])
        self.assertAlmostEqual(himmelblau(x), 0.0)

    def test_multiple(self):
        himmelblau = Himmelblau()
        x = np.array([[3, 2], [3, 2]])
        assert_array_almost_equal(himmelblau(x), np.array([0.0, 0.0]))


class TestThreeHumpCamel(unittest.TestCase):
    def test(self):
        thc = ThreeHumpCamel()
        x = np.array([0, 0])
        self.assertAlmostEqual(thc(x), 0.0)

    def test_multiple(self):
        thc = ThreeHumpCamel()
        x = np.array([[0, 0], [0, 0]])
        assert_array_almost_equal(thc(x), np.array([0.0, 0.0]))


class TestEasom(unittest.TestCase):
    def test(self):
        easom = Easom()
        x = np.array([np.pi, np.pi])
        self.assertAlmostEqual(easom(x), -1.0)

    def test_multiple(self):
        easom = Easom()
        x = np.array([[np.pi, np.pi], [np.pi, np.pi]])
        assert_array_almost_equal(easom(x), np.array([-1.0, -1.0]))


class TestCrossInTray(unittest.TestCase):
    def test(self):
        cit = CrossInTray()
        x = np.array([1.34941, 1.34941])
        self.assertAlmostEqual(cit(x), -2.06261, places=5)

    def test_multiple(self):
        cit = CrossInTray()
        x = np.array([[1.34941, 1.34941], [1.34941, 1.34941]])
        assert_array_almost_equal(cit(x), np.array([-2.06261, -2.06261]),
                                  decimal=5)


class TestEggholder(unittest.TestCase):
    def test(self):
        eggholder = Eggholder()
        x = np.array([512, 404.2319])
        self.assertAlmostEqual(eggholder(x), -959.6407, places=4)

    def test_multiple(self):
        eggholder = Eggholder()
        x = np.array([[512, 404.2319], [512, 404.2319]])
        assert_array_almost_equal(eggholder(x),
                                  np.array([-959.6407, -959.6407]), decimal=4)


class TestMcCormick(unittest.TestCase):
    def test(self):
        mccormick = McCormick()
        x = np.array([-0.54719, -1.54719])
        self.assertAlmostEqual(mccormick(x), -1.9133, places=3)

    def test_multiple(self):
        mccormick = McCormick()
        x = np.array([[-0.54719, -1.54719], [-0.54719, -1.54719]])
        assert_array_almost_equal(mccormick(x), np.array([-1.9133, -1.9133]),
                                  decimal=4)


class TestSchafferN2(unittest.TestCase):
    def test(self):
        schaffer_n2 = SchafferN2()
        x = np.array([0, 0])
        self.assertAlmostEqual(schaffer_n2(x), 0.0)

    def test_multiple(self):
        schaffer_n2 = SchafferN2()
        x = np.array([[0, 0], [0, 0]])
        assert_array_almost_equal(schaffer_n2(x), np.array([0.0, 0.0]))


class TestSchafferN4(unittest.TestCase):
    def test(self):
        schaffer_n4 = SchafferN4()
        x = np.array([0, 1.25313])
        self.assertAlmostEqual(schaffer_n4(x), 0.292579, places=3)

    def test_multiple(self):
        schaffer_n4 = SchafferN4()
        x = np.array([[0, 1.25313], [0, 1.25313]])
        assert_array_almost_equal(schaffer_n4(x),
                                  np.array([0.292579, 0.292579]), decimal=3)


class TestDropWave(unittest.TestCase):
    def test(self):
        drop_wave = DropWave()
        x = np.array([0, 0])
        self.assertAlmostEqual(drop_wave(x), -1.0)

    def test_multiple(self):
        drop_wave = DropWave()
        x = np.array([[0, 0], [0, 0]])
        assert_array_almost_equal(drop_wave(x), np.array([-1.0, -1.0]))


class TestRastrigin(unittest.TestCase):
    def test(self):
        rastrigin = Rastrigin(dim=2)
        x = np.array([0, 0])
        self.assertAlmostEqual(rastrigin(x), 0.0)

    def test_multiple(self):
        rastrigin = Rastrigin(dim=2)
        x = np.array([[0, 0], [0.0, 0.0]])
        assert_array_almost_equal(rastrigin(x), np.array([0.0, 0.0]))

    def test_high_dim(self):
        rastrigin = Rastrigin(dim=100)
        x = np.zeros(100)
        self.assertAlmostEqual(rastrigin(x), 0.0)


class TestRosenbrock(unittest.TestCase):
    def test(self):
        rosenbrock = Rosenbrock(dim=2)
        x = np.array([1, 1])
        self.assertAlmostEqual(rosenbrock(x), 0.0)

    def test_multiple(self):
        rosenbrock = Rosenbrock(dim=2)
        x = np.array([[1, 1], [1.0, 1.0]])
        assert_array_almost_equal(rosenbrock(x), np.array([0.0, 0.0]))

    def test_high_dim(self):
        rosenbrock = Rosenbrock(dim=100)
        x = np.ones(100)
        self.assertAlmostEqual(rosenbrock(x), 0.0)


class TestAckley(unittest.TestCase):
    def test(self):
        ackley = Ackley(dim=2)
        x = np.array([0, 0])
        self.assertAlmostEqual(ackley(x), 0.0)

    def test_multiple(self):
        ackley = Ackley(dim=2)
        x = np.array([[0, 0], [0.0, 0.0]])
        assert_array_almost_equal(ackley(x), np.array([0.0, 0.0]))

    def test_high_dim(self):
        ackley = Ackley(dim=100)
        x = np.zeros(100)
        self.assertAlmostEqual(ackley(x), 0.0)


class TestSphere(unittest.TestCase):
    def test(self):
        sphere = Sphere(dim=2)
        x = np.array([0, 0])
        self.assertAlmostEqual(sphere(x), 0.0)

    def test_multiple(self):
        sphere = Sphere(dim=2)
        x = np.array([[0, 0], [0.0, 0.0]])
        assert_array_almost_equal(sphere(x), np.array([0.0, 0.0]))

    def test_high_dim(self):
        sphere = Sphere(dim=100)
        x = np.zeros(100)
        self.assertAlmostEqual(sphere(x), 0.0)


class TestStyblinskiTang(unittest.TestCase):
    def test(self):
        styblinski_tang = StyblinskiTang(dim=2)
        x = np.array([-2.903534, -2.903534])
        self.assertAlmostEqual(styblinski_tang(x), -78.332, places=3)

    def test_multiple(self):
        styblinski_tang = StyblinskiTang(dim=2)
        x = np.array([[-2.903534, -2.903534], [-2.903534, -2.903534]])
        assert_array_almost_equal(styblinski_tang(x),
                                  np.array([-78.332, -78.332]), decimal=3)

    def test_high_dim(self):
        styblinski_tang = StyblinskiTang(dim=100)
        x = np.ones(100) * -2.903534
        self.assertAlmostEqual(styblinski_tang(x),
                               styblinski_tang.get_optimum_value(), places=3)


class TestGriewank(unittest.TestCase):
    def test(self):
        griewank = Griewank(dim=2)
        x = np.array([0, 0])
        self.assertAlmostEqual(griewank(x), 0.0)

    def test_multiple(self):
        griewank = Griewank(dim=2)
        x = np.array([[0, 0], [0.0, 0.0]])
        assert_array_almost_equal(griewank(x), np.array([0.0, 0.0]))

    def test_high_dim(self):
        griewank = Griewank(dim=100)
        x = np.zeros(100)
        self.assertAlmostEqual(griewank(x), 0.0)


class TestLevy(unittest.TestCase):
    def test(self):
        levy = Levy(dim=2)
        x = np.array([1, 1])
        self.assertAlmostEqual(levy(x), 0.0)

    def test_multiple(self):
        levy = Levy(dim=2)
        x = np.array([[1, 1], [1.0, 1.0]])
        assert_array_almost_equal(levy(x), np.array([0.0, 0.0]))

    def test_high_dim(self):
        levy = Levy(dim=100)
        x = np.ones(100)
        self.assertAlmostEqual(levy(x), 0.0)


# TODO test DixonPrice
# TODO test Trid
# TODO test Schwefel
# TODO test Shubert

class TestAll(unittest.TestCase):
    def test(self):
        for name, obj in inspect.getmembers(sys.modules[__name__]):
            if inspect.isclass(obj) and issubclass(obj, unittest.TestCase):
                suite = unittest.TestLoader().loadTestsFromTestCase(obj)
                unittest.TextTestRunner(verbosity=2).run(suite)



import multiprocessing
_numba_cache = True  # Set to True/False to enable/disable Numba cache
_numba_parallel = False  # Set to True/False to enable/disable Numba parallel
_numba_nopython = True  # Set to True/False to enable/disable Numba nopython
_numba_nogil = True  # Set to True/False to enable/disable Numba nogil
# Set to the number of threads to use for Numba
_numba_num_threads = multiprocessing.cpu_count()
_disable_jit = True  # Set to True/False to enable/disable Numba JIT



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(
    name='open-metaoptim',
    version='0.0.14',
    description='A Python package for metaheuristic optimisation.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/harrymyburgh/metaoptim',
    project_urls={
        'Bug Tracker': 'https://github.com/harrymyburgh/metaoptim/issues'
    },
    author='Harry Phillip Myburgh',
    author_email='harry.myburgh@gmail.com',
    py_modules=['metaoptim'],
    package_dir={'': 'src'},
    packages=[
        'metaoptim',
        'metaoptim.pso',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.9',
    install_requires=[
        'numpy >= 1.23.0',
        'numba >= 0.56.0',
        'tqdm >= 4.64.0',
    ],
    extras_require={
        'dev': [
            'pytest >= 7.2.0',
            'check-manifest >= 0.49',
            'twine >= 4.0.2',
            'pycodestyle >= 2.10.0',
        ]
    },
    include_package_data=True,
)



